unit WinLayerCore;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls, Math,
  ExtCtrls, PngImage, WinLayerFS, WinLayerConts;

const
  MaxPixelCountA = MaxInt div SizeOf(TRGBQuad);
  MaxPixelCount = MaxInt div SizeOf(TRGBTriple);

type
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..MaxPixelCount-1] of TRGBTriple;
  PRGBAArray = ^TRGBAArray;
  TRGBAArray = array[0..MaxPixelCountA-1] of TRGBQuad;

  TDIBSection = class // only for 32bit data
  private
    FWidth: Integer;
    FHeight: Integer;
    FBitCount: Integer;
    FBits: Pointer;
    FCanvas: TCanvas;
    FDC: HDC;
    FBitmap, FOldBitmap: HBITMAP;
    FBitmapInfo: PBitmapInfo;
    function GetScanline(Index: Integer): Pointer;
    function GetPixels(X, Y: Integer): Pointer;
    function GetSize: Integer;
    function GetRect: TRect;
  public
    constructor Create;
    destructor Destroy; override;
    procedure FreeResources;

    function SetSize(AWidth, AHeight, ABitCount: Integer): Boolean;
    procedure Assign(Src: TDIBSection);

    procedure LoadFromStream(Stream: TStream);
    procedure LoadFromFile(FileName: String);
    procedure SaveToStream(Stream: TStream);
    procedure SaveToFile(FileName: String);

    procedure PremultiplyRect(Rect: TRect);
    procedure DrawDIB(DestRect: TRect; Src: TDIBSection; SrcRect: TRect);
    procedure DrawPNG(DestRect: TRect; Src: TPNGObject; SrcRect: TRect; AlphaTransparent: Boolean);
    procedure DrawTo(DC: HDC; Dest: TRect; Src: TRect);

    property Width: Integer read FWidth;
    property Height: Integer read FHeight;
    property BitCount: Integer read FBitCount;
    property Scanline[Index: Integer]: Pointer read GetScanline;
    property Pixels[X, Y: Integer]: Pointer read GetPixels;
    property Bits: Pointer read FBits;
    property Size: Integer read GetSize;
    property Rect: TRect read GetRect;
    property Canvas: TCanvas read FCanvas;
    property BitmapInfo: PBitmapInfo read FBitmapInfo;
  end;

  TWinLayerLibType = (ltImage, ltColor, ltInt, ltFont);

  TWinLayerLibItem = class(TCollectionItem)
  private
    FDataType: TWinLayerLibType;
    FData: Pointer;
    FName: String;

    function GetColor: TColor;
    procedure SetColor(Value: TColor);
    function GetInt: String;
    procedure SetInt(Value: String);
    function GetImage: TPNGObject;
    procedure SetImage(Value: TPNGObject);
    function GetFont: TFont;
    procedure SetFont(Value: TFont);
    procedure SetDataType(Value: TWinLayerLibType);
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    function Int(Index: Integer): Integer;
    function Rect: TRect;

    property Data: Pointer read FData write FData;
  published
    property Name: String read FName write FName nodefault;
    property DataType: TWinLayerLibType read FDataType write SetDataType default ltImage;
    property AsColor: TColor read GetColor write SetColor nodefault;
    property AsInt: String read GetInt write SetInt nodefault;
    property AsImage: TPNGObject read GetImage write SetImage nodefault;
    property AsFont: TFont read GetFont write SetFont nodefault;
  end;

  TWinLayerLibItems = class(TCollection)
  private
    FOwner: TComponent;

    function GetItem(Index: Integer): TWinLayerLibItem;
    procedure SetItem(Index: Integer; Value: TWinLayerLibItem);
    function Get(Name: String): TWinLayerLibItem;
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(AOwner: TComponent);

    procedure Changed;
    function Add: TWinLayerLibItem;

    property Items[Index: Integer]: TWinLayerLibItem read GetItem write SetItem;
    property GetItems[Name: String]: TWinLayerLibItem read Get; default;
  end;

  TWinLayerLib = class(TComponent)
  private
    FItems: TWinLayerLibItems;
    FAutoOpen: Boolean;
    FOpened: Boolean;
    FSkinFile: TFileName;

    procedure SetItems(Value: TWinLayerLibItems);
  protected
    procedure ReadState(Reader: TReader); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Open; overload;
    procedure Open(FileName: String); overload;
    procedure Save; overload;
    procedure Save(FileName: String); overload;
    procedure Update;
  published
    property SkinFile: TFileName read FSkinFile write FSkinFile;
    property Items: TWinLayerLibItems read FItems write SetItems;
    property AutoOpen: Boolean read FAutoOpen write FAutoOpen default False;
    property Opened: Boolean read FOpened write FOpened;
  end;

  TWinLayerFormButtonState = (bsNormal, bsHover, bsPush, bsDisable);

  TWinLayerForm = class;
  TWinLayerCustomControl = class(TCustomControl)
  private
    FWinLayerForm: TWinLayerForm;
    FVisible: Boolean;
    FAlignment: TAlignment;

    procedure SetWinLayerForm(Value: TWinLayerForm);
    procedure SetVisible(Value: Boolean);
  protected
    procedure ReadState(Reader: TReader); override;
    procedure SetAlignment(Value: TAlignment); virtual;

    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Paint; override;
    procedure Draw;
    function GetDIB: TDIBSection; virtual;
  published
    property WinLayerForm: TWinLayerForm read FWinLayerForm write SetWinLayerForm;
    property Align;
    property Enabled;
    property Visible: Boolean read FVisible write SetVisible default True;
  end;

  TWinLayerHint = class(THintWindow)
  private
  public
  end;

  TWinLayerForm = class(TComponent)
  private
    FWndProc: TWndMethod;
    FInitialized: Boolean;
    FInitialization: Boolean;
    FZSize: TSize;
    FZPoint: TPoint;
    FZBF: TBlendFunction;
    FZTopLeft: TPoint;
    FDIBWnd: TDIBSection;
    FDIB: TDIBSection;
    FAutoInit: Boolean;
    FLib: TWinLayerLib;
    FDownPoint: TPoint;
    FDown: Boolean;
    FWndSize: Integer;
    FBorderStyle: TBorderStyle;
    FPainting: Boolean;
    FButtonClose: TWinLayerFormButtonState;
    FButtonMinimize: TWinLayerFormButtonState;
    FButtonMaximize: TWinLayerFormButtonState;
    FButtonDown: Integer;
    FControls: TList;
    FVCLControls: TList;
    FDIBWndIsReady: Boolean;
    FIsFormActive: Boolean;

    function GetControlsCount: Integer;
    function GetControl(Index: Integer): TWinLayerCustomControl;
  protected
    procedure DoWndProc(var Message: TMessage);
    procedure DoMouseDown(Message: TWMMOUSE);
    procedure DoMouseUp(Message: TWMMOUSE);
    procedure DoMouseMove(Message: TWMMOUSE);
    procedure DoSize(Message: TWMSIZE);

    procedure ReadState(Reader: TReader); override;
    function GetButtonRect(BorderIcon: TBorderIcon): TRect;
    procedure DrawButton(BorderIcons: TBorderIcons);

    procedure PrepareDIBWnd;
    procedure DrawCaption;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Init;
    procedure Finit;
    procedure BeginPaint;
    procedure EndPaint;
    procedure Update;
    procedure Draw;

    procedure AddVCLControl(Control: TControl);
    procedure RemoveVCLControl(Control: TControl);
    procedure RemoveVCLControls;
    procedure PaintVCLControl(Control: TControl);

    procedure PaintControl(Index: Integer); overload;
    procedure PaintControl(Control: TWinLayerCustomControl); overload;
    procedure RemoveControl(Index: Integer); overload;
    procedure RemoveControl(Control: TWinLayerCustomControl); overload;
    procedure RemoveControls;
    procedure AddControl(Control: TWinLayerCustomControl);
    procedure InsertControl(Index: Integer; Control: TWinLayerCustomControl);

    property Initialized: Boolean read FInitialized;
    property Painting: Boolean read FPainting;
    property BorderStyle: TBorderStyle read FBorderStyle;
    property ControlsCount: Integer read GetControlsCount;
    property Controls[Index: Integer]: TWinLayerCustomControl read GetControl;
    property Surface: TDIBSection read FDIB;
    property SurfaceWindow: TDIBSection read FDIBWnd;
  published
    property AutoInit: Boolean read FAutoInit write FAutoInit default True;
    property WLibrary: TWinLayerLib read FLib write FLib nodefault;
  end;

procedure Register;

function PosFrom(Substr, S: String; Start: Integer): Integer;
function CalcInt(Str: String; StrParams, IntParams: array of const): Integer;
function mRect(Left, Top, Width, Height: Integer): TRect;
procedure DrawTitlePNGTo(Dest: TDIBSection; DestRect: TRect; Src: TPNGObject;
  SrcRect: TRect; AlphaTransparent: Boolean);

implementation

procedure Register;
begin
  RegisterComponents('WinLayerLib', [TWinLayerForm, TWinLayerLib]);
end;

function PosFrom(Substr, S: String; Start: Integer): Integer;
var
  p, psub: PCHAR;
  i: Integer;
begin
  Result := 0;
  p := PCHAR(Integer(Pointer(S)) + Start - 1);
  psub := PCHAR(Substr);
  i := Start;
  while (p^ <> #0) do
  begin                      
    if (p^ = psub^) then
    begin
      inc(psub);
      if (psub^ = #0) then
      begin
        Result := i - Length(Substr) + 1;
        Break;
      end;
    end else
    if (psub <> PCHAR(Substr)) then
      psub := PCHAR(Substr);
    inc(p);
    inc(i);
  end;
end;

function CalcInt(Str: String; StrParams, IntParams: array of const): Integer;
var
  Tmp: String;

procedure Replace(SubStr: String; Val: Integer);
var
  i, start: Integer;
begin
  start := 1;
  while True do
  begin
    i := PosFrom(SubStr, Tmp, start);
    if (i = 0) then
      Break;
    Delete(Tmp, i, Length(SubStr));
    Insert(IntToStr(Val), Tmp, i);
    start := i + 1;
  end;
end;

var
  start: Integer;
  i, e: Integer;
  n: Real;
  v: Real;
  Stack: array of String;
  p: PCHAR;
  Proc: Integer;
begin
  Result := 0;
  if (Length(Str) = 0) then Exit;

  Tmp := Str;
  Tmp := LowerCase(Tmp);

  for i:= low(StrParams) to high(StrParams) do
    Replace(StrParams[i].VPChar, IntParams[i].VInteger);

  start := 1;
  i := 1;
  p := PCHAR(Tmp);

  while (p^ <> #0) do
  begin
    if (p^ = #32) then
    begin
      Delete(Tmp, i, 1);
      p := PCHAR(Integer(PCHAR(Tmp)) + i - 1);
      Continue;
    end;
    if (p^ in ['+', '-', '/', '*']) then
    begin
      if (i - start > 0) then
      begin
        SetLength(Stack, Length(Stack) + 1);
        Stack[Length(Stack) - 1] := Copy(Tmp, start, i - start);
      end;
      SetLength(Stack, Length(Stack) + 1);
      Stack[Length(Stack) - 1] := p^;
      start := i + 1;
    end;
    inc(i);
    inc(p);
  end;
  if (i - start > 0) then
  begin
    SetLength(Stack, Length(Stack) + 1);
    Stack[Length(Stack) - 1] := Copy(Tmp, start, i - start);
  end;

  v := 0;
  Proc := 0;
  for i:= 0 to Length(Stack) - 1 do
  if (Stack[i] = '+') then Proc := 0 else
  if (Stack[i] = '-') then Proc := 1 else
  if (Stack[i] = '/') then Proc := 2 else
  if (Stack[i] = '*') then Proc := 3 else
  begin
    val(Stack[i], n, e);
    if (e = 0) then
    case Proc of
      0:
        v := v + n;
      1:
        v := v - n;
      2:
        v := v / n;
      3:
        v := v * n;
    end;
  end;
  SetLength(Stack, 0);

  Result := Trunc(v);
end;

function mRect(Left, Top, Width, Height: Integer): TRect;
begin
  Result.Left := Left;
  Result.Top := Top;
  Result.Right := Left + Width;
  Result.Bottom := Top + Height;
end;

procedure DrawTitlePNGTo(Dest: TDIBSection; DestRect: TRect; Src: TPNGObject;
  SrcRect: TRect; AlphaTransparent: Boolean);
var
  x, y: Integer;
  ws, hs: Integer;
begin
  if (Src.Height = 0) or (Src.Width = 0) then Exit;
  y := DestRect.Top;
  ws := SrcRect.Right - SrcRect.Left;
  hs := SrcRect.Bottom - SrcRect.Top;

  while (y + hs < DestRect.Bottom) do
  begin
    x := DestRect.Left;
    while (x + ws < DestRect.Right) do
    begin
      Dest.DrawPNG(Rect(x, y, x + ws, y + hs), Src, SrcRect, AlphaTransparent);
      inc(x, ws);
    end;
    Dest.DrawPNG(Rect(x, y, DestRect.Right, y + hs), Src,
      Rect(SrcRect.Left, SrcRect.Top,
      SrcRect.Left + DestRect.Right - x, SrcRect.Bottom), AlphaTransparent);
    inc(y, hs);
  end;

  x := DestRect.Left;
  while (x + ws < DestRect.Right) do
  begin
    Dest.DrawPNG(Rect(x, y, x + ws, DestRect.Bottom), Src,
      Rect(SrcRect.Left, SrcRect.Top, SrcRect.Right,
      SrcRect.Top + DestRect.Bottom - y), AlphaTransparent);
    inc(x, ws);
  end;
  Dest.DrawPNG(Rect(x, y, DestRect.Right, DestRect.Bottom), Src,
    Rect(SrcRect.Left, SrcRect.Top,
    SrcRect.Left + DestRect.Right - x,
    SrcRect.Top + DestRect.Bottom - y), AlphaTransparent);
end;   

{ TDIBSection }

constructor TDIBSection.Create;
begin
  inherited;
  FWidth := 0;
  FHeight := 0;
  FBitCount := 0;
  FBits := NIL;
  FCanvas := NIL;
end;

destructor TDIBSection.Destroy;
begin
  FreeResources;
  inherited;
end;

procedure TDIBSection.Assign(Src: TDIBSection);
begin
  SetSize(Src.Width, Src.Height, Src.BitCount);
  DrawDIB(Rect, Src, Src.Rect);
end;

procedure TDIBSection.DrawDIB(DestRect: TRect; Src: TDIBSection; SrcRect: TRect);
var
  wd, ws, hd, hs: Integer;
  FactorX, FactorY: Double;
  i, j: Integer;
  RowSrc, RowDest: PRGBAArray;
  BytesPerRowDest, BytesPerRowSrc: Integer;
begin
  wd := DestRect.Right - DestRect.Left;
  hd := DestRect.Bottom - DestRect.Top;
  ws := SrcRect.Right - SrcRect.Left;
  hs := SrcRect.Bottom - SrcRect.Top;
  if (ws = 0) or (wd = 0) then FactorX := 1 else FactorX := wd / ws;
  if (hs = 0) or (hd = 0) then FactorY := 1 else FactorY := hd / hs;
  BytesPerRowDest := FWidth * 4;
  BytesPerRowSrc := Src.Width * 4;
  
  Longint(RowDest) := Longint(FBits) + (FHeight - DestRect.Top - 1) *
    BytesPerRowDest + DestRect.Left * 4;

  for i:= 0 to hd - 1 do
  begin
    Longint(RowSrc) := Longint(Src.Bits) + (Src.Height - SrcRect.Top - Trunc(i / FactorY) - 1) *
      BytesPerRowSrc + SrcRect.Left * 4;
    for j:= 0 to wd - 1 do
      RowDest[j] := RowSrc[Trunc(j / FactorX)];
    dec(Longint(RowDest), BytesPerRowDest);
  end;
end;
  
procedure TDIBSection.DrawPNG(DestRect: TRect; Src: TPNGObject;
  SrcRect: TRect; AlphaTransparent: Boolean);
var
  wd, ws, hd, hs: Integer;
  FactorX, FactorY: Double;
  ImageAlpha, ImageSource: PByteArray;
  ImageDest: PRGBAArray;
  i, j, i2, j2: Integer;
  BytesPerRowDest: Integer;
begin
  wd := DestRect.Right - DestRect.Left;
  hd := DestRect.Bottom - DestRect.Top;
  ws := SrcRect.Right - SrcRect.Left;
  hs := SrcRect.Bottom - SrcRect.Top;
  if (ws = 0) or (wd = 0) then FactorX := 1 else FactorX := wd / ws;
  if (hs = 0) or (hd = 0) then FactorY := 1 else FactorY := hd / hs;
  BytesPerRowDest := FWidth * 4;

  Integer(ImageDest) := Integer(FBits) + (FHeight - DestRect.Top - 1) *
    BytesPerRowDest + DestRect.Left * 4;

  for i:= 0 to hd - 1 do
  begin
    i2 := Trunc(i / FactorY);
    if (SrcRect.Top + i2 >= Integer(Src.Header.Height)) then Break;
    Integer(ImageSource) := Integer(Src.Scanline[SrcRect.Top + i2]) + SrcRect.Left * 3;
    Integer(ImageAlpha) := Integer(Src.AlphaScanline[SrcRect.Top + i2]) + SrcRect.Left;

    for j:= 0 to wd - 1 do
    begin
      j2 := Trunc(j / FactorX);

      if (ImageAlpha[j2] = 0) then
      begin
        if AlphaTransparent then
        ImageDest[j].rgbReserved := 0;
      end else
      if (ImageAlpha[j2] = MAXBYTE) then
      begin
        ImageDest[j].rgbBlue := ImageSource[j2 * 3];
        ImageDest[j].rgbGreen := ImageSource[j2 * 3 + 1];
        ImageDest[j].rgbRed := ImageSource[j2 * 3 + 2];
        if AlphaTransparent then
        ImageDest[j].rgbReserved := MAXBYTE;
      end else
      begin
        ImageDest[j].rgbBlue := (MAXBYTE + ImageSource[j2 * 3] * ImageAlpha[j2] +
          ImageDest[j].rgbBlue * (not ImageAlpha[j2])) shr 8;
        ImageDest[j].rgbGreen := (MAXBYTE + ImageSource[j2 * 3 + 1] * ImageAlpha[j2] +
          ImageDest[j].rgbGreen * (not ImageAlpha[j2])) shr 8;
        ImageDest[j].rgbRed := (MAXBYTE + ImageSource[j2 * 3 + 2] * ImageAlpha[j2] +
          ImageDest[j].rgbRed * (not ImageAlpha[j2])) shr 8;
        if AlphaTransparent then
        ImageDest[j].rgbReserved := ImageAlpha[j2];
      end;
    end;
    
    dec(Integer(ImageDest), BytesPerRowDest);
  end;
end;

function TDIBSection.GetRect: TRect;
begin
  Result.Left := 0;
  Result.Top := 0;
  Result.Right := FWidth;
  Result.Bottom := FHeight;
end;

function TDIBSection.GetPixels(X, Y: Integer): Pointer;
begin
  Result := NIL;
  if (FBits <> NIL) then
    Result := Pointer(Integer(FBits) + (FWidth * Y + X) * (FBitCount div 8)); 
end;

function TDIBSection.GetScanline(Index: Integer): Pointer;
begin
  Result := NIL;
  if (FBits <> NIL) then
    Result := Pointer(Integer(FBits) + FWidth * (FBitCount div 8) * (FHeight - Index - 1));
end;

function TDIBSection.GetSize: Integer;
begin
  Result := FWidth * FHeight * (FBitCount div 8);
end;

procedure TDIBSection.FreeResources;
begin
  if (FCanvas <> NIL) then
  begin
    FCanvas.Handle := 0;
    FCanvas.Free;
    FCanvas := NIL;

    SelectObject(FDC, FOldBitmap);
    DeleteObject(FBitmap);
    DeleteObject(FDC);
    dispose(FBitmapInfo);
  end;
end;

function TDIBSection.SetSize(AWidth, AHeight, ABitCount: Integer): Boolean;
begin
  Result := False;
  if not (ABitCount in [32]) or (AWidth <= 0) or (AHeight <= 0) then
    Exit;

  FreeResources;

  FWidth := AWidth;
  FHeight := AHeight;
  FBitCount := ABitCount;

  new(FBitmapInfo);
  with FBitmapInfo^.bmiHeader do
  begin
    biSize := SizeOf(TBitmapInfoHeader);
    biWidth := FWidth;
    biHeight := FHeight;
    biPlanes := 1;
    biBitCount := FBitCount;
    biCompression := BI_RGB;
    biSizeImage := Size;
    biXPelsPerMeter := 0;
    biYPelsPerMeter := 0;
    biClrUsed := 0;
    biClrImportant := 0;
  end;
  FDC := CreateCompatibleDC(0);
  FBitmap := CreateDIBSection(FDC, FBitmapInfo^, DIB_RGB_COLORS, FBits, 0, 0);
  FOldBitmap := SelectObject(FDC, FBitmap);
  FCanvas := TCanvas.Create;
  FCanvas.Handle := FDC;

  Result := True;
end;

procedure TDIBSection.DrawTo(DC: HDC; Dest, Src: TRect);
begin
  StretchDIBits(DC, Dest.Left, Dest.Top, Dest.Right - Dest.Left, Dest.Bottom - Dest.Top,
    Src.Left, Src.Top, Src.Right - Src.Left, Src.Bottom - Src.Top, FBits, FBitmapInfo^,
    DIB_RGB_COLORS, SRCCOPY);
end;

procedure TDIBSection.PremultiplyRect(Rect: TRect);
var
  i, j: integer;
  Row: PRGBAArray;
begin
  for i := Rect.Top to Rect.Bottom - 1 do
  begin
    Row := Scanline[i];
    for j := Rect.Left to Rect.Right - 1 do
    if (Row[j].rgbReserved = 0) then
      Row[j].rgbReserved := MAXBYTE;
  end;
end;

procedure TDIBSection.LoadFromFile(FileName: String);
var
  F: TFileStream;
begin
  F := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(F);
  finally
    F.Free;
  end;
end;

procedure TDIBSection.LoadFromStream(Stream: TStream);
var
  bitFileHeader: TBitmapFileHeader;
  bitInfo: TBitmapInfo;
begin
  Stream.Read(bitFileHeader, SizeOf(TBitmapFileHeader));
  Stream.Read(bitInfo, SizeOf(TBitmapInfo));
  if (bitInfo.bmiHeader.biBitCount <> 32) then Exit;
  SetSize(bitInfo.bmiHeader.biWidth, bitInfo.bmiHeader.biHeight, 32);
  Stream.Seek(bitFileHeader.bfOffBits, 0);
  Stream.Read(FBits^, Size); 
end;

procedure TDIBSection.SaveToFile(FileName: String);
var
  F: TFileStream;
begin
  F := TFileStream.Create(FileName, fmCreate or fmOpenWrite);
  try
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

procedure TDIBSection.SaveToStream(Stream: TStream);
var
  bitFileHeader: TBitmapFileHeader;
begin
  bitFileHeader.bfType := $4D42;
  bitFileHeader.bfSize := SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfo) + Size;
  bitFileHeader.bfReserved1 := 0;
  bitFileHeader.bfReserved2 := 0;
  bitFileHeader.bfOffBits := SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfo);

  Stream.Write(bitFileHeader, SizeOf(TBitmapFileHeader));
  Stream.Write(FBitmapInfo^, SizeOf(TBitmapInfo));
  Stream.Write(FBits^, Size); 
end;

{ TWinLayerForm }

type
  PVCLControlData = ^TVCLControlData;
  TVCLControlData = packed record
    Form: TWinLayerForm;
    Control: TWinControl;
    WndProc: Pointer;
  end;

var
  AllVCLControls: TList;

function VCLControlsIndexOf(Wnd: HWND): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i:= 0 to AllVCLControls.Count - 1 do
  if (PVCLControlData(AllVCLControls[i])^.Control.Handle = Wnd) then
  begin
    Result := i;
    Break;
  end;
end;
  
function VCLWndProc(Wnd: HWND; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  i: Integer;
  p: PVCLControlData;
begin
  i := VCLControlsIndexOf(Wnd);
  if (i < 0) then
  begin
    Result := 1;
    Exit;
  end;
  p := AllVCLControls[i];
  Result := CallWindowProc(p^.WndProc, Wnd, Msg, wParam, lParam);
  case Msg of
    WM_PAINT, WM_NCPAINT, WM_CHAR:
    begin        
      p^.Form.PaintVCLControl(p^.Control);
      p^.Form.Update;
    end;
  end; 
end;

constructor TWinLayerForm.Create(AOwner: TComponent);
begin
  inherited;
  FAutoInit := True;
  FLib := NIL;
  FButtonDown := -4;
  FControls := TList.Create;
  FVCLControls := TList.Create;
  FDIBWndIsReady := False;
  FInitialized := False;
  FInitialization := False;
end;

destructor TWinLayerForm.Destroy;
begin
  Finit;
  RemoveControls;
  FControls.Free;
  RemoveVCLControls;
  FVCLControls.Free;
  inherited;
end;

procedure TWinLayerForm.Init;
const
  ButtonState: array[Boolean] of TWinLayerFormButtonState = (bsDisable, bsNormal);
var
  i: Integer;
begin          
  if not FInitialized and not FInitialization and
    not (csDesigning in ComponentState) and Assigned(FLib) then
  begin
    FInitialization := True;

    FBorderStyle := TForm(Owner).BorderStyle;
    TForm(Owner).BorderStyle := bsNone;

    FWndProc := TForm(Owner).WindowProc;
    TForm(Owner).WindowProc := DoWndProc;
                        
    SetWindowLong(TForm(Owner).Handle, GWL_EXSTYLE,
      GetWindowLong(TForm(Owner).Handle, GWL_EXSTYLE) or WS_EX_LAYERED);

    FButtonClose := ButtonState[(biSystemMenu in TForm(Owner).BorderIcons) and
      (FBorderStyle in [bsSingle, bsSizeable, bsDialog, bsToolWindow, bsSizeToolWin])];
    FButtonMinimize := ButtonState[(biMinimize in TForm(Owner).BorderIcons) and
      (FBorderStyle in [bsSingle, bsSizeable])];
    FButtonMaximize := ButtonState[(biMaximize in TForm(Owner).BorderIcons) and
      (FBorderStyle in [bsSingle, bsSizeable])];       

    FDIBWnd := TDIBSection.Create;
    FDIBWnd.SetSize(TForm(Owner).Width, TForm(Owner).Height, 32);
    FDIB := TDIBSection.Create;
    FDIB.SetSize(TForm(Owner).Width, TForm(Owner).Height, 32);
    FDIBWndIsReady := False;
                                   
    FZSize.cx := TForm(Owner).Width;
    FZSize.cy := TForm(Owner).Height;
    FZPoint := Point(0, 0);
            
    with FZBF do
    begin
      BlendOp := AC_SRC_OVER;
      BlendFlags := 0;
      AlphaFormat := AC_SRC_ALPHA;
      SourceConstantAlpha := MAXBYTE;
    end;

    for i:= 0 to TForm(Owner).ControlCount - 1 do
    begin  
      if FLib.Opened then
      with TForm(Owner).Controls[i] do
      begin
        if (Left < 0) then Left := 0;
        if (Top < 0) then Top := 0;
        SetBounds(Left + FLib.Items[WinLayer_window_src_wnd_active_left].Int(2),
          Top + FLib.Items[WinLayer_window_src_wnd_active_top].Int(3), Width, Height);
      end;
      AddVCLControl(TForm(Owner).Controls[i]);
    end;
    
    FInitialized := True;
    FInitialization := False;

    if FLib.Opened then
      Draw;
  end;
end;

procedure TWinLayerForm.Finit;
var
  i: Integer;
begin
  if FInitialized then
  begin
    FInitialized := False;
    if Assigned(TForm(Owner)) then
    begin
      for i:= 0 to TForm(Owner).ControlCount - 1 do
      with TForm(Owner).Controls[i] do
      begin
        SetBounds(Left - FLib.Items[WinLayer_window_src_wnd_active_left].Int(2),
          Top - FLib.Items[WinLayer_window_src_wnd_active_top].Int(3), Width, Height);
      end;

      TForm(Owner).WindowProc := FWndProc;
      TForm(Owner).BorderStyle := FBorderStyle;
      SetWindowLong(TForm(Owner).Handle, GWL_EXSTYLE,
        GetWindowLong(TForm(Owner).Handle, GWL_EXSTYLE) and not WS_EX_LAYERED);
    end;
    FDIBWnd.Free;
    FDIB.Free;
  end;
end;

const
  _ButtonClose: array[TWinLayerFormButtonState] of String =
    (WinLayer_window_src_button_close_normal, WinLayer_window_src_button_close_hover,
    WinLayer_window_src_button_close_push, WinLayer_window_src_button_close_disable);
    
  _ButtonMinimize: array[TWinLayerFormButtonState] of String =
    (WinLayer_window_src_button_minimize_normal, WinLayer_window_src_button_minimize_hover,
    WinLayer_window_src_button_minimize_push, WinLayer_window_src_button_minimize_disable);

  _ButtonMaximize: array[TWinLayerFormButtonState] of String =
    (WinLayer_window_src_button_maximize_normal, WinLayer_window_src_button_maximize_hover,
    WinLayer_window_src_button_maximize_push, WinLayer_window_src_button_maximize_disable);

  window_lefttop = 0;
  window_righttop = 1;
  window_leftbottom = 2;
  window_rightbottom = 3;
  window_left = 4;
  window_right = 5;
  window_top = 6;
  window_bottom = 7;
  window_bckg = 8;
  _WindowState: array[window_lefttop..window_bckg, Boolean] of String =
    ((WinLayer_window_src_wnd_inactive_lefttop,WinLayer_window_src_wnd_active_lefttop),
    (WinLayer_window_src_wnd_inactive_righttop,WinLayer_window_src_wnd_active_righttop),
    (WinLayer_window_src_wnd_inactive_leftbottom,WinLayer_window_src_wnd_active_leftbottom),
    (WinLayer_window_src_wnd_inactive_rightbottom,WinLayer_window_src_wnd_active_rightbottom),
    (WinLayer_window_src_wnd_inactive_left,WinLayer_window_src_wnd_active_left),
    (WinLayer_window_src_wnd_inactive_right,WinLayer_window_src_wnd_active_right),
    (WinLayer_window_src_wnd_inactive_top,WinLayer_window_src_wnd_active_top),
    (WinLayer_window_src_wnd_inactive_bottom,WinLayer_window_src_wnd_active_bottom),
    (WinLayer_window_src_wnd_inactive_bckg,WinLayer_window_src_wnd_active_bckg));

function TWinLayerForm.GetButtonRect(BorderIcon: TBorderIcon): TRect;
begin
  case BorderIcon of
    biSystemMenu:
    begin
      Result.Left := FLib.Items[WinLayer_window_pos_button_close].Int(0);
      Result.Top := FLib.Items[WinLayer_window_pos_button_close].Int(1);
      Result.Right := Result.Left + FLib.Items[_ButtonClose[FButtonClose]].Int(2);
      Result.Bottom := Result.Top + FLib.Items[_ButtonClose[FButtonClose]].Int(3);
    end;
    biMinimize:
    begin
      Result.Left := FLib.Items[WinLayer_window_pos_button_minimize].Int(0);
      Result.Top := FLib.Items[WinLayer_window_pos_button_minimize].Int(1);
      Result.Right := Result.Left + FLib.Items[_ButtonMinimize[FButtonMinimize]].Int(2);
      Result.Bottom := Result.Top + FLib.Items[_ButtonMinimize[FButtonMinimize]].Int(3);
    end;
    biMaximize:
    begin
      Result.Left := FLib.Items[WinLayer_window_pos_button_maximize].Int(0);
      Result.Top := FLib.Items[WinLayer_window_pos_button_maximize].Int(1);
      Result.Right := Result.Left + FLib.Items[_ButtonMaximize[FButtonMaximize]].Int(2);
      Result.Bottom := Result.Top + FLib.Items[_ButtonMaximize[FButtonMaximize]].Int(3);
    end;
  end;
end;

procedure TWinLayerForm.DrawButton(BorderIcons: TBorderIcons);
begin
  if (BorderIcons = []) then Exit;

  if (biSystemMenu in BorderIcons) then
  begin
    FDIB.DrawDIB(GetButtonRect(biSystemMenu), FDIBWnd, GetButtonRect(biSystemMenu));
    FDIB.DrawPNG(GetButtonRect(biSystemMenu),
      FLib.Items[WinLayer_src_image].AsImage,
      FLib.Items[_ButtonClose[FButtonClose]].Rect,
      False);
  end;

  if (biMinimize in BorderIcons) then
  begin
    FDIB.DrawDIB(GetButtonRect(biMinimize), FDIBWnd, GetButtonRect(biMinimize));
    FDIB.DrawPNG(GetButtonRect(biMinimize),
      FLib.Items[WinLayer_src_image].AsImage,
      FLib.Items[_ButtonMinimize[FButtonMinimize]].Rect,
      False);
  end;

  if (biMaximize in BorderIcons) then
  begin
    FDIB.DrawDIB(GetButtonRect(biMaximize), FDIBWnd, GetButtonRect(biMaximize));
    FDIB.DrawPNG(GetButtonRect(biMaximize),
      FLib.Items[WinLayer_src_image].AsImage,
      FLib.Items[_ButtonMaximize[FButtonMaximize]].Rect,
      False);
  end;
end;

procedure TWinLayerForm.PrepareDIBWnd;
begin
  if not FInitialized then Exit;
  
  FDIBWnd.DrawPNG(Rect(0, 0, FLib.Items[_WindowState[window_lefttop, FIsFormActive]].Int(2),
    FLib.Items[_WindowState[window_lefttop, FIsFormActive]].Int(3)),
    FLib.Items[WinLayer_src_image].AsImage,
    FLib.Items[_WindowState[window_lefttop, FIsFormActive]].Rect, True);

  FDIBWnd.DrawPNG(Rect(FLib.Items[_WindowState[window_lefttop, FIsFormActive]].Int(2), 0,
      TForm(Owner).Width - FLib.Items[_WindowState[window_righttop, FIsFormActive]].Int(2),
      FLib.Items[_WindowState[window_top, FIsFormActive]].Int(3)),
    FLib.Items[WinLayer_src_image].AsImage,
    FLib.Items[_WindowState[window_top, FIsFormActive]].Rect, True);

  FDIBWnd.DrawPNG(Rect(TForm(Owner).Width -
    FLib.Items[_WindowState[window_righttop,FIsFormActive]].Int(2), 0,
    TForm(Owner).Width, FLib.Items[_WindowState[window_righttop, FIsFormActive]].Int(3)),
    FLib.Items[WinLayer_src_image].AsImage,
    FLib.Items[_WindowState[window_righttop, FIsFormActive]].Rect, True);

  FDIBWnd.DrawPNG(Rect(0, FLib.Items[_WindowState[window_lefttop, FIsFormActive]].Int(3),
      FLib.Items[_WindowState[window_left, FIsFormActive]].Int(2),
      TForm(Owner).Height - FLib.Items[_WindowState[window_leftbottom, FIsFormActive]].Int(3)),
    FLib.Items[WinLayer_src_image].AsImage,
    FLib.Items[_WindowState[window_left, FIsFormActive]].Rect, True);

  FDIBWnd.DrawPNG(Rect(TForm(Owner).Width -
    FLib.Items[_WindowState[window_right, FIsFormActive]].Int(2),
    FLib.Items[_WindowState[window_righttop, FIsFormActive]].Int(3),
    TForm(Owner).Width,
    TForm(Owner).Height - FLib.Items[_WindowState[window_rightbottom, FIsFormActive]].Int(3)),
    FLib.Items[WinLayer_src_image].AsImage,
    FLib.Items[_WindowState[window_right, FIsFormActive]].Rect, True);

  FDIBWnd.DrawPNG(Rect(0, TForm(Owner).Height -
    FLib.Items[_WindowState[window_leftbottom, FIsFormActive]].Int(3),
    FLib.Items[_WindowState[window_leftbottom, FIsFormActive]].Int(2), TForm(Owner).Height),
    FLib.Items[WinLayer_src_image].AsImage,
    FLib.Items[_WindowState[window_leftbottom, FIsFormActive]].Rect, True);

  FDIBWnd.DrawPNG(Rect(TForm(Owner).Width -
    FLib.Items[_WindowState[window_rightbottom, FIsFormActive]].Int(2),
    TForm(Owner).Height - FLib.Items[_WindowState[window_rightbottom, FIsFormActive]].Int(3),
    TForm(Owner).Width, TForm(Owner).Height),
    FLib.Items[WinLayer_src_image].AsImage,
    FLib.Items[_WindowState[window_rightbottom, FIsFormActive]].Rect, True);

  FDIBWnd.DrawPNG(Rect(FLib.Items[_WindowState[window_leftbottom, FIsFormActive]].Int(2),
      TForm(Owner).Height - FLib.Items[_WindowState[window_bottom, FIsFormActive]].Int(3),
      TForm(Owner).Width - FLib.Items[_WindowState[window_rightbottom, FIsFormActive]].Int(2),
      TForm(Owner).Height),
    FLib.Items[WinLayer_src_image].AsImage,
    FLib.Items[_WindowState[window_bottom, FIsFormActive]].Rect, True);

  DrawTitlePNGTo(FDIBWnd, Rect(FLib.Items[_WindowState[window_left, FIsFormActive]].Int(2),
    FLib.Items[_WindowState[window_top, FIsFormActive]].Int(3),
    TForm(Owner).Width - FLib.Items[_WindowState[window_right, FIsFormActive]].Int(2),
    TForm(Owner).Height - FLib.Items[_WindowState[window_bottom, FIsFormActive]].Int(3)),
    FLib.Items[WinLayer_src_image].AsImage, FLib.Items[_WindowState[window_bckg, FIsFormActive]].Rect,
    True);
                    
  FDIBWndIsReady := True;
end;

procedure TWinLayerForm.DrawCaption;
var
  lpRect: TRect;
  uFormat: DWORD;
begin
  lpRect := FLib.Items[WinLayer_window_pos_caption].Rect;
  uFormat := DT_VCENTER or DT_SINGLELINE or DT_WORDBREAK or DT_WORD_ELLIPSIS;
  case FLib.Items[WinLayer_window_pos_caption_align].Int(0) of
    1: // left
      uFormat := uFormat or DT_LEFT;
    2: // center
      uFormat := uFormat or DT_CENTER;
    3: // right
      uFormat := uFormat or DT_RIGHT;
  end;
  FDIB.DrawDIB(lpRect, FDIBWnd, lpRect);
  FDIB.Canvas.Brush.Style := bsClear;
  FDIB.Canvas.Font.Assign(FLib.Items[WinLayer_font_window_caption].AsFont);
  DrawText(FDIB.Canvas.Handle, PCHAR(TForm(Owner).Caption),
    Length(TForm(Owner).Caption), lpRect, uFormat);
  FDIB.Canvas.Brush.Style := bsSolid;
  FDIB.PremultiplyRect(lpRect);
end;

procedure TWinLayerForm.BeginPaint;
begin
  if FPainting or not FInitialized then Exit;
  FPainting := True;

  if not FDIBWndIsReady then
    PrepareDIBWnd;

  FDIB.Assign(FDIBWnd);

  DrawCaption;
  DrawButton([biSystemMenu, biMinimize, biMaximize]);
end;

procedure TWinLayerForm.EndPaint;
var
  i: Integer;
begin
  if FPainting then
  begin
    for i:= 0 to FControls.Count - 1 do PaintControl(i);
    for i:= 0 to FVCLControls.Count - 1 do PaintVCLControl(PVCLControlData(FVCLControls[i])^.Control);
    FPainting := False;
  end;
end;

procedure TWinLayerForm.Update;
begin
  if FInitialized then
  begin
    FZTopLeft := Point(TForm(Owner).Left, TForm(Owner).Top);
    UpdateLayeredWindow(TForm(Owner).Handle, 0, @FZTopLeft, @FZSize, FDIB.Canvas.Handle,
      @FZPoint, 0, @FZBF, ULW_ALPHA);
  end;    
end;

procedure TWinLayerForm.PaintVCLControl(Control: TControl);
var
  r: TRect;
begin
  if not FInitialized or not (Control is TWinControl) or not Control.Visible or
    (Control is TWinLayerCustomControl) then Exit;
                                     
  r := Control.BoundsRect;
  if (r.Left < FLib.Items[_WindowState[window_left, FIsFormActive]].Int(2)) then
  begin
    r.Left := FLib.Items[_WindowState[window_left, FIsFormActive]].Int(2);
    r.Right := r.Left + Control.Width;
  end;
  if (r.Top < FLib.Items[_WindowState[window_top, FIsFormActive]].Int(3)) then
  begin
    r.Top := FLib.Items[_WindowState[window_top, FIsFormActive]].Int(3);
    r.Bottom := r.Top + Control.Height;
  end;
  if (r.Right >= TForm(Owner).Width - FLib.Items[_WindowState[window_right, FIsFormActive]].Int(2)) then
    r.Right := TForm(Owner).Width - FLib.Items[_WindowState[window_right, FIsFormActive]].Int(2);
  if (r.Bottom >= TForm(Owner).Height - FLib.Items[_WindowState[window_bottom, FIsFormActive]].Int(3)) then
    r.Bottom := TForm(Owner).Height - FLib.Items[_WindowState[window_bottom, FIsFormActive]].Int(3);
    
  TWinControl(Control).PaintTo(FDIB.Canvas, Control.Left, Control.Top);
  FDIB.PremultiplyRect(Control.BoundsRect);
end;

procedure TWinLayerForm.ReadState(Reader: TReader);
begin
  inherited;
  if FAutoInit and not (csDesigning in ComponentState) then
    Init;
end;

procedure TWinLayerForm.DoMouseUp(Message: TWMMouse);
var
  p: TPoint;
begin
  FDown := False;
  p := SmallPointToPoint(Message.Pos);
  FButtonDown := -4;

  if (FButtonClose = bsPush) then
  begin
    if PtInRect(GetButtonRect(biSystemMenu), p) then
    begin
      TForm(Owner).Close;
      FButtonClose := bsHover;
    end else
      FButtonClose := bsNormal;
    DrawButton([biSystemMenu]);
    Update;
  end else       
  if (FButtonMinimize = bsPush) then
  begin
    if PtInRect(GetButtonRect(biMinimize), p) then
    begin
      if (TForm(Owner) = Application.MainForm) then
        Application.Minimize
      else    
        TForm(Owner).WindowState := wsMinimized;
    end;
    FButtonMinimize := bsNormal;
    DrawButton([biMinimize]);
    Update;
  end
  else                
  if (FButtonMaximize = bsPush) and PtInRect(GetButtonRect(biMaximize), p) then
  begin
    if PtInRect(GetButtonRect(biMaximize), p) then
    begin
      if (TForm(Owner).WindowState = wsMaximized) then
        TForm(Owner).WindowState := wsNormal
      else
        TForm(Owner).WindowState := wsMaximized;
    end;
    FButtonMaximize := bsNormal;
    DrawButton([biMaximize]);
    Update;
  end;
end;

procedure TWinLayerForm.DoMouseDown(Message: TWMMouse);
var
  p: TPoint;

function DoButton(BorderIcon: TBorderIcon; var Button: TWinLayerFormButtonState): Boolean;
begin
  Result := False;
  if (Button <> bsDisable) then
  if PtInRect(GetButtonRect(BorderIcon), p) then
  begin
    if (Button <> bsPush) then
    begin
      Button := bsPush;
      DrawButton([BorderIcon]);
      Update;
    end;
    Result := True;
  end;
end;

begin
  FButtonDown := -4;
  p := SmallPointToPoint(Message.Pos);

  if DoButton(biSystemMenu, FButtonClose) then FButtonDown := -1 else
  if DoButton(biMinimize, FButtonMinimize) then FButtonDown := -2 else
  if DoButton(biMaximize, FButtonMaximize) then FButtonDown := -3 else
  if (TForm(Owner).WindowState = wsNormal) and (FWndSize = 0) and
    PtInRect(FLib.Items[WinLayer_window_size_caption].Rect, p) then
  begin
    ReleaseCapture;
    TForm(Owner).Perform(WM_SYSCOMMAND, $F012, 0);
  end else
  begin
    FDownPoint := p;
    FDown := True;
  end;
end;

procedure TWinLayerForm.DoMouseMove(Message: TWMMouse);
var
  p: TPoint;
  Cursor: TCursor;

procedure DoRight;
begin
  if (TForm(Owner).Width - FDownPoint.X + p.X >= FLib.Items[WinLayer_window_size_min].Int(2)) then
  begin
    with TForm(Owner) do
    SetBounds(
      Left,
      Top,
      Width - FDownPoint.X + p.X,
      Height);
    FDownPoint.X := p.X;
  end;
end;

procedure DoLeft;
begin
  if (TForm(Owner).Width + FDownPoint.X - p.X >= FLib.Items[WinLayer_window_size_min].Int(2)) then
  begin
    with TForm(Owner) do
    SetBounds(
      Left - FDownPoint.X + p.X,
      Top,
      Width + FDownPoint.X - p.X,
      Height);
  end;
end;

procedure DoTop;
begin
  if (TForm(Owner).Height + FDownPoint.Y - p.Y >= FLib.Items[WinLayer_window_size_min].Int(3)) then
  begin
    with TForm(Owner) do
    SetBounds(
      Left,
      Top - FDownPoint.Y + p.Y,
      Width,
      Height + FDownPoint.Y - p.Y);
  end;
end;

procedure DoBottom;
begin
  if (TForm(Owner).Height - FDownPoint.Y + p.Y >= FLib.Items[WinLayer_window_size_min].Int(3)) then
  begin
    with TForm(Owner) do
    SetBounds(
      Left,
      Top,
      Width,
      Height - FDownPoint.Y + p.Y);
    FDownPoint.Y := p.Y;
  end;
end;

procedure DoButton(BorderIcon: TBorderIcon; var Button: TWinLayerFormButtonState;
  IdButton: Integer);
begin
  if (Button <> bsDisable) then
  if PtInRect(GetButtonRect(BorderIcon), p) then
  begin
    if (FButtonDown = -4) and (Button <> bsHover) then
    begin
      Button := bsHover;
      DrawButton([BorderIcon]);
      Update;
    end else
    if (FButtonDown = IdButton) and (Button <> bsPush) then
    begin
      Button := bsPush;
      DrawButton([BorderIcon]);
      Update;
    end;
  end else
  if (Button <> bsNormal) then
  begin
    Button := bsNormal;
    DrawButton([BorderIcon]);
    Update;
  end;
end;

begin
  p := SmallPointToPoint(Message.Pos);

  DoButton(biSystemMenu, FButtonClose, -1);
  DoButton(biMinimize, FButtonMinimize, -2);
  DoButton(biMaximize, FButtonMaximize, -3);

  if (FBorderStyle = bsSizeable) then
  if FDown then
  begin
    case FWndSize of
      1: // right, bottom
      begin
        DoRight;
        DoBottom;
      end;
      2: // left, top
      begin
        DoLeft;
        DoTop;
      end;
      3: // left, bottom
      begin
        DoLeft;
        DoBottom;
      end;
      4: // right, top
      begin
        DoRight;
        DoTop;
      end;
      5: // top;
        DoTop;
      6: // bottom
        DoBottom;
      7: // right
        DoRight;
      8: // left
        DoLeft;
    end;
  end else
  begin
    Cursor := crDefault;
    FWndSize := 0;

    if (TForm(Owner).WindowState = wsNormal) then

    if PtInRect(FLib.Items[WinLayer_window_size_rightbottom].Rect, p) then
    begin
      Cursor := crSizeNWSE;
      FWndSize := 1;
    end else

    if PtInRect(FLib.Items[WinLayer_window_size_lefttop].Rect, p) then
    begin
      Cursor := crSizeNWSE;
      FWndSize := 2;
    end else

    if PtInRect(FLib.Items[WinLayer_window_size_leftbottom].Rect, p) then
    begin
      Cursor := crSizeNESW;
      FWndSize := 3;
    end else
    
    if PtInRect(FLib.Items[WinLayer_window_size_righttop].Rect, p) then
    begin
      Cursor := crSizeNESW;
      FWndSize := 4;
    end else

    if PtInRect(FLib.Items[WinLayer_window_size_top].Rect, p) then
    begin
      Cursor := crSizeNS;
      FWndSize := 5;
    end else

    if PtInRect(FLib.Items[WinLayer_window_size_bottom].Rect, p) then
    begin
      Cursor := crSizeNS;
      FWndSize := 6;
    end else

    if PtInRect(FLib.Items[WinLayer_window_size_right].Rect, p) then
    begin
      Cursor := crSizeWE;
      FWndSize := 7;
    end else

    if PtInRect(FLib.Items[WinLayer_window_size_left].Rect, p) then
    begin
      Cursor := crSizeWE;
      FWndSize := 8;
    end;

    TForm(Owner).Cursor := Cursor;
  end;
end;

procedure TWinLayerForm.DoSize(Message: TWMSIZE);
begin
  if FInitialized then
  begin
    FZSize.cx := TForm(Owner).Width;
    FZSize.cy := TForm(Owner).Height;
    FDIBWnd.SetSize(TForm(Owner).Width, TForm(Owner).Height, 32);
    FDIB.SetSize(TForm(Owner).Width, TForm(Owner).Height, 32);
    FDIBWndIsReady := False;
    Draw;
  end;
end;

procedure TWinLayerForm.Draw;
begin
  BeginPaint;
  EndPaint;
  Update;
end;

function TWinLayerForm.GetControl(Index: Integer): TWinLayerCustomControl;
begin
  Result := FControls[Index];
end;

function TWinLayerForm.GetControlsCount: Integer;
begin
  Result := FControls.Count;
end;

procedure TWinLayerForm.RemoveControl(Index: Integer);
begin
  TWinLayerCustomControl(FControls[Index]).Free;
  FControls.Delete(Index);
end;

procedure TWinLayerForm.RemoveControl(Control: TWinLayerCustomControl);
begin
  RemoveControl(FControls.IndexOf(Control));
end;

procedure TWinLayerForm.RemoveControls;
var
  i: Integer;
begin
  for i:= FControls.Count - 1 downto 0 do
    TWinLayerCustomControl(FControls[i]).Free;
end;

procedure TWinLayerForm.InsertControl(Index: Integer; Control: TWinLayerCustomControl);
begin
  Control.WinLayerForm := Self;
  FControls.Insert(Index, Control);
end;

procedure TWinLayerForm.AddControl(Control: TWinLayerCustomControl);
begin
  InsertControl(FControls.Count, Control);
end;

procedure TWinLayerForm.PaintControl(Index: Integer);
begin
  PaintControl(TWinLayerCustomControl(FControls[Index]));
end;

procedure TWinLayerForm.PaintControl(Control: TWinLayerCustomControl);
var
  DIB: TDIBSection;
  r: TRect;
begin
  if not FInitialized or not Control.Visible then Exit;
  DIB := Control.GetDIB;
  if Assigned(DIB) then
  begin
    r := Control.BoundsRect;
    if (r.Left < FLib.Items[_WindowState[window_left, FIsFormActive]].Int(2)) then
    begin
      r.Left := FLib.Items[_WindowState[window_left, FIsFormActive]].Int(2);
      r.Right := r.Left + Control.Width;
    end;
    if (r.Top < FLib.Items[_WindowState[window_top, FIsFormActive]].Int(3)) then
    begin
      r.Top := FLib.Items[_WindowState[window_top, FIsFormActive]].Int(3);
      r.Bottom := r.Top + Control.Height;
    end;
    if (r.Right >= TForm(Owner).Width - FLib.Items[_WindowState[window_right, FIsFormActive]].Int(2)) then
      r.Right := TForm(Owner).Width - FLib.Items[_WindowState[window_right, FIsFormActive]].Int(2);
    if (r.Bottom >= TForm(Owner).Height - FLib.Items[_WindowState[window_bottom, FIsFormActive]].Int(3)) then
      r.Bottom := TForm(Owner).Height - FLib.Items[_WindowState[window_bottom, FIsFormActive]].Int(3);
    FDIB.DrawDIB(r, DIB, Rect(0, 0, r.Right - r.Left, r.Bottom - r.Top));
    DIB.Free;
  end;
end;

procedure TWinLayerForm.AddVCLControl(Control: TControl);
{var
  p: PVCLControlData;   }
begin
 {if (Control is TWinControl) then
  begin
    new(p);
    p^.Form := Self;
    p^.Control := TWinControl(Control);
    p^.WndProc := Pointer(GetWindowLong(p^.Control.Handle, GWL_WNDPROC));
    //SetWindowLong(p^.Control.Handle, GWL_WNDPROC, Integer(@VCLWndProc));
    //FVCLControls.Add(p);
    AllVCLControls.Add(p);   
  end;    }
end;    

procedure TWinLayerForm.RemoveVCLControl(Control: TControl);
var
  i: Integer;                 
begin
  for i:= 0 to FVCLControls.Count - 1 do
  if (PVCLControlData(FVCLControls[i])^.Control = Control) then
  begin
    SetWindowLong(PVCLControlData(FVCLControls[i])^.Control.Handle, GWL_WNDPROC,
      Integer(PVCLControlData(FVCLControls[i])^.WndProc));
    AllVCLControls.Remove(PVCLControlData(FVCLControls[i]));
    dispose(PVCLControlData(FVCLControls[i]));
    FVCLControls.Delete(i);
    Break;
  end;
end;                 

procedure TWinLayerForm.RemoveVCLControls;
var
  i: Integer;
begin
  for i:= FVCLControls.Count - 1 downto 0 do
    RemoveVCLControl(FVCLControls[i]);
end;

procedure TWinLayerForm.DoWndProc(var Message: TMessage);
const
  ButtonState: array[Boolean] of TWinLayerFormButtonState = (bsDisable, bsNormal);
var
  Msg: TMessage;
begin
  Msg := Message;

  case Msg.Msg of
    CM_RECREATEWND:
    if FInitialized and not FInitialization then
    begin
      FInitialized := False;
      if Assigned(TForm(Owner)) then
      begin
        TForm(Owner).WindowProc := FWndProc;
        TForm(Owner).Cursor := crDefault;
      end;
      FWndSize := 0;
      FDown := False;
      FDIBWnd.Free;
      FDIB.Free;
    end;
  end;

  if Assigned(FWndProc) then
    FWndProc(Message);
  
  case Msg.Msg of
    WM_LBUTTONDOWN:
      DoMouseDown(TWMMouse(Msg));

    WM_LBUTTONUP:
      DoMouseUp(TWMMouse(Msg));

    WM_MOUSEMOVE:
      DoMouseMove(TWMMouse(Msg));

    WM_SIZE:
      DoSize(TWMSize(Msg));

    WM_NCACTIVATE:
    begin
      FIsFormActive := Boolean(Msg.WParam);
      Draw;
    end;

    CM_TEXTCHANGED:
    begin
      DrawCaption;
      Update;
    end;

    CM_RECREATEWND:
      Init;
  end;
end;

{ TWinLayerLibItem }

constructor TWinLayerLibItem.Create(Collection: TCollection);
begin
  inherited;
  FDataType := ltImage;
  FData := TPNGObject.Create;
end;

destructor TWinLayerLibItem.Destroy;
begin
  case FDataType of
    ltInt:
      dispose(PString(FData));
    ltColor:
      dispose(PColor(FData));
    ltImage:
      TPNGObject(FData).Free;
    ltFont:
      TFont(FData).Free;
  end;
  inherited;
end;

function TWinLayerLibItem.GetColor: TColor;
begin
  Result := clNone;
  if (FDataType = ltColor) then
    Result := PColor(FData)^;
end;

function TWinLayerLibItem.GetDisplayName: string;
begin
  Result := FName;
end;

function TWinLayerLibItem.GetFont: TFont;
begin
  Result := NIL;
  if (FDataType = ltFont) then
    Result := FData;
end;

function TWinLayerLibItem.GetImage: TPNGObject;
begin
  Result := NIL;
  if (FDataType = ltImage) then
    Result := FData;
end;

function TWinLayerLibItem.GetInt: String;
begin
  Result := '';
  if (FDataType = ltInt) then
    Result := PString(FData)^;
end;

procedure TWinLayerLibItem.SetColor(Value: TColor);
begin
  if (FDataType = ltColor) then
    PColor(FData)^ := Value;
end;

procedure TWinLayerLibItem.SetDataType(Value: TWinLayerLibType);
var
  p: PColor;
  s: PString;
begin
  if (FDataType <> Value) then
  begin
    case FDataType of
      ltColor:
        dispose(PColor(FData));
      ltInt:
        dispose(PString(FData));
      ltImage:
        TPNGObject(FData).Free;
      ltFont:
        TFont(FData).Free;
    end;
    FDataType := Value;
    case FDataType of
      ltInt:
        begin
          new(s);
          s^ := '';
          FData := s;
        end;
      ltColor:
        begin
          new(p);
          p^ := clNone;
          FData := p;
        end;
      ltImage:
        FData := TPNGObject.Create;
      ltFont:
        FData := TFont.Create;
    end;
  end;
end;

procedure TWinLayerLibItem.SetFont(Value: TFont);
begin
  if (FDataType = ltFont) then
    TFont(FData).Assign(Value);
end;

procedure TWinLayerLibItem.SetImage(Value: TPNGObject);
begin
  if (FDataType = ltImage) then
    TPNGObject(FData).Assign(Value);
end;

procedure TWinLayerLibItem.SetInt(Value: String);
begin
  if (FDataType = ltInt) then
    PString(FData)^ := Value;
end;

function TWinLayerLibItem.Int(Index: Integer): Integer;

function ExtractPart: String;
var
  p: PCHAR;
  i, n, m: Integer;
begin
  Result := '';
  p := PCHAR(PString(FData)^);
  n := 1;
  m := 1;
  i := 0;
  while (p^ <> #0) do
  begin
    if (p^ = ',') then
    begin
      if (i = Index) then
      begin
        Result := Copy(PString(FData)^, n, m - n);
        Exit;
      end else
      begin
        n := m + 1;
        inc(i);                  
      end;
    end;
    inc(p);
    inc(m);
  end;
  if (i = Index) then
    Result := Copy(PString(FData)^, n, m - n);
end;

begin
  Result := 0;
  if (FDataType = ltInt) and (Index in [0..3]) then
    Result := CalcInt(ExtractPart,
      ['$form.width', '$form.height'],
      [TForm(TWinLayerLib(Collection.Owner).Owner).Width,
      TForm(TWinLayerLib(Collection.Owner).Owner).Height])
end;

function TWinLayerLibItem.Rect: TRect;
begin
  Result.Left := Int(0);
  Result.Top := Int(1);
  Result.Right := Result.Left + Int(2);
  Result.Bottom := Result.Top + Int(3);
end;

{ TWinLayerLibItems }

constructor TWinLayerLibItems.Create;
begin
  inherited Create(TWinLayerLibItem);
  FOwner := AOwner;
end;

function TWinLayerLibItems.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

function TWinLayerLibItems.Add: TWinLayerLibItem;
begin
  Result := TWinLayerLibItem(inherited Add);
end;

function TWinLayerLibItems.GetItem(Index: Integer): TWinLayerLibItem;
begin
  Result := TWinLayerLibItem(inherited GetItem(Index));
end;

procedure TWinLayerLibItems.SetItem(Index: Integer; Value: TWinLayerLibItem);
begin
  inherited SetItem(Index, Value);
end;

procedure TWinLayerLibItems.Changed;
begin
  inherited Changed;
end;

function TWinLayerLibItems.Get(Name: String): TWinLayerLibItem;
var
  i: Integer;
begin
  Result := NIL;
  for i:= 0 to Count - 1 do
  if (CompareStr(Items[i].Name, Name) = 0) then
  begin
    Result := Items[i];
    Break;
  end;
end;

{ TWinLayerLib }

constructor TWinLayerLib.Create(AOwner: TComponent);
begin
  inherited;
  FItems := TWinLayerLibItems.Create(Self);
  FAutoOpen := False;
  FOpened := False;
end;

destructor TWinLayerLib.Destroy;
begin
  FItems.Free;
  inherited;
end;

procedure TWinLayerLib.SetItems(Value: TWinLayerLibItems);
begin
  FItems.Assign(Value); 
end;

procedure TWinLayerLib.Open(FileName: String);
const
  WinLayerType: array[WinLayerTypeImage..WinLayerTypeFont] of TWinLayerLibType =
    (ltImage, ltColor, ltInt, ltFont);
var
  fs: TWinLayerFS;
  i: Integer;
begin
  FItems.BeginUpdate;
  FItems.Clear;
  fs := TWinLayerFS.Create(FileName, True);
  for i:= 0 to fs.Count - 1 do
  with FItems.Add do
  begin
    Name := fs.Item[i].dwName;
    DataType := WinLayerType[fs.Item[i].dwType];
    case DataType of
      ltInt:
        AsInt := PString(fs.Item[i].dwData)^;
      ltColor:
        AsColor := PWORD(fs.Item[i].dwData)^;
      ltImage:
        begin
          fs.Prepare(i, True);
          AsImage := TPNGObject(fs.Item[i].dwData);
        end;
      ltFont:
        begin
          AsFont.Name := PWinLayerFSFont(fs.Item[i].dwData)^.Name;
          AsFont.Size := PWinLayerFSFont(fs.Item[i].dwData)^.Size;
          AsFont.Color := PWinLayerFSFont(fs.Item[i].dwData)^.Color;
          AsFont.Style := BinToFontStyles(PWinLayerFSFont(fs.Item[i].dwData)^.Style);
        end;
    end;
  end;
  fs.Free;
  FOpened := True;
  FItems.EndUpdate;
  Update;
end;

procedure TWinLayerLib.Open;
begin
  Open(FSkinFile);
end;

procedure TWinLayerLib.Save(FileName: String);
const
  WinLayerType: array[TWinLayerLibType] of Byte =
    (WinLayerTypeImage, WinLayerTypeColor, WinLayerTypeInt, WinLayerTypeFont);
var
  fs: TWinLayerFS;
  i: Integer;
begin
  fs := TWinLayerFS.Create(FileName, False);
  for i:= 0 to FItems.Count - 1 do
  with FItems.Items[i] do
  case DataType of
    ltInt:
      fs.AddInt(Name, AsInt);
    ltColor:
      fs.AddColor(Name, AsColor);
    ltImage:
      fs.AddImage(Name, AsImage);
    ltFont:
      fs.AddFont(Name, AsFont.Name, AsFont.Size, AsFont.Color, AsFont.Style);
  end;
  fs.Free;
end;

procedure TWinLayerLib.Save;
begin
  Save(FSkinFile);
end;

procedure TWinLayerLib.ReadState(Reader: TReader);
begin
  inherited;
  if FAutoOpen then
    Open;
end;

procedure TWinLayerLib.Update;
var
  i, j: Integer;
begin
  for i:= 0 to Screen.FormCount - 1 do
  for j:= 0 to Screen.Forms[i].ComponentCount - 1 do
  if (Screen.Forms[i].Components[j] is TWinLayerForm) then
    TWinLayerForm(Screen.Forms[i].Components[j]).Draw;
end;

{ TWinLayerCustomControl }

constructor TWinLayerCustomControl.Create(AOwner: TComponent);
begin
  inherited;
  FWinLayerForm := NIL;
  FVisible := True;
  FAlignment := taLeftJustify;
end;

destructor TWinLayerCustomControl.Destroy;
begin
  SetWinLayerForm(NIL);
  inherited;
end;

procedure TWinLayerCustomControl.SetVisible(Value: Boolean);
begin
  if (Value <> FVisible) then
  begin
    FVisible := Value;
    if not FVisible then
    begin
      FWinLayerForm.Surface.DrawDIB(BoundsRect, FWinLayerForm.SurfaceWindow, BoundsRect);
      FWinLayerForm.Update;
    end else
      Draw;
  end;
end;

procedure TWinLayerCustomControl.Draw;
begin
  if Visible and Assigned(FWinLayerForm) and FWinLayerForm.Initialized then
  begin     
    FWinLayerForm.PaintControl(Self);
    FWinLayerForm.Update;
  end;
end;

procedure TWinLayerCustomControl.SetWinLayerForm(Value: TWinLayerForm);
begin
  if (Value <> FWinLayerForm) then
  begin
    if Assigned(FWinLayerForm) then
      FWinLayerForm.FControls.Remove(Self); 
    FWinLayerForm := Value;   
    if Assigned(FWinLayerForm) then
      FWinLayerForm.AddControl(Self);
  end;
end;

procedure TWinLayerCustomControl.ReadState(Reader: TReader);
begin
  inherited;
  if Assigned(FWinLayerForm) then
    FWinLayerForm.AddControl(Self);
end;

procedure TWinLayerCustomControl.Paint;
begin
  if (csDesigning in ComponentState) then
  begin
    Canvas.Brush.Color := clWhite;
    Canvas.Pen.Color := clBtnShadow;
    Canvas.Pen.Style := psDash;
    Canvas.Rectangle(Canvas.ClipRect);
    Canvas.Font.Assign(Font); 
    Canvas.TextOut((Width - Canvas.TextWidth(Name)) div 2,
      (Height - Canvas.TextHeight('W')) div 2, Name);
  end;
end;

function TWinLayerCustomControl.GetDIB: TDIBSection;
begin
  Result := NIL;
end;

procedure TWinLayerCustomControl.SetAlignment(Value: TAlignment);
begin
  if (Value <> FAlignment) then
  begin
    FAlignment := Value;
    Draw;
  end;
end;

initialization
  AllVCLControls := TList.Create;

finalization
  AllVCLControls.Free;

end.
