unit WinLayerFS;

interface

uses
  Windows, SysUtils, Classes, Graphics, CRC, PngImage;

const
  WinLayerTypeImage = 1;
  WinLayerTypeColor = 2;
  WinLayerTypeInt = 3;
  WinLayerTypeFont = 4;

  WinLayerFS_Font_Bold = $0001;
  WinLayerFS_Font_Italic = $0002;
  WinLayerFS_Font_Underline = $0004;
  WinLayerFS_Font_Strikeout = $0008;

type
  TWinLayerFSHeader = packed record
    dwHead: array[0..13] of CHAR; // "WinLayer SKIN"
    dwSize: DWORD; // File size, without Header
    dwDescOffset: DWORD;
    dwNameOffset: DWORD;
    dwDataOffset: DWORD;
    dwCRC: DWORD;
  end;

  TWinLayerFSHeaderDescItem = packed record
    dwFlag: Byte; // only 0xFF
    dwType: Byte;
    dwNameLen: WORD; // length of file's name ( Unicode - WChar - 0x2 )
    dwNameOffset: DWORD;
    dwSize: DWORD; // data size
    dwOffset: DWORD;
  end;

  TWinLayerFSItem = packed record
    dwName: String;
    dwType: Byte;
    dwState: Boolean;
    dwOffset: DWORD;
    dwSize: DWORD;
    dwData: Pointer;
  end;

  PWinLayerFSFont = ^TWinLayerFSFont;
  TWinLayerFSFont = packed record
    Name: String;
    Size: Integer;
    Color: DWORD;
    Style: WORD;
  end;

  TWinLayerFS = class
  private
    FFileName: String;
    FAsOpen: Boolean;
    FFile: HFILE;
    FItems: array of TWinLayerFSItem;

    function GetCount: Integer;
    function GetItem(Index: Integer): TWinLayerFSItem;
    procedure SetItem(Index: Integer; Value: TWinLayerFSItem);

    procedure Read;
    procedure Write;
  public
    constructor Create(FileName: String; AsOpen: Boolean);
    destructor Destroy; override;

    procedure AddImage(Name: String; Image: TPNGObject); overload;
    procedure AddImage(Name, FileName: String); overload;
    procedure AddColor(Name: String; Color: DWORD);
    procedure AddInt(Name: String; Value: String);
    procedure AddFont(Name, FontName: String; Size: Integer; Color: TColor;
      Style: TFontStyles);

    procedure Delete(Index: Integer);
    procedure Clear;

    procedure Prepare(Index: Integer; State: Boolean);

    property Count: Integer read GetCount;
    property Item[Index: Integer]: TWinLayerFSItem read GetItem write SetItem;
    property FileName: String read FFileName;
  end;

  function FontStylesToBin(Value: TFontStyles): WORD;
  function BinToFontStyles(Value: WORD): TFontStyles;

implementation

function FontStylesToBin(Value: TFontStyles): WORD;
begin
  Result := 0;
  if (fsBold in Value) then Result := Result or WinLayerFS_Font_Bold;
  if (fsItalic in Value) then Result := Result or WinLayerFS_Font_Italic;
  if (fsUnderline in Value) then Result := Result or WinLayerFS_Font_Underline;
  if (fsStrikeOut in Value) then Result := Result or WinLayerFS_Font_Strikeout;
end;

function BinToFontStyles(Value: WORD): TFontStyles;
begin
  Result := [];
  if (Value and WinLayerFS_Font_Bold = WinLayerFS_Font_Bold) then
    Result := Result + [fsBold];
  if (Value and WinLayerFS_Font_Italic = WinLayerFS_Font_Italic) then
    Result := Result + [fsItalic];
  if (Value and WinLayerFS_Font_Underline = WinLayerFS_Font_Underline) then
    Result := Result + [fsUnderline];
  if (Value and WinLayerFS_Font_Strikeout = WinLayerFS_Font_Strikeout) then
    Result := Result + [fsStrikeOut];
end;

{ TWinLayerFS }

constructor TWinLayerFS.Create(FileName: String; AsOpen: Boolean);
begin
  inherited Create;
  FFileName := FileName;
  FAsOpen := AsOpen;
  FFile := INVALID_HANDLE_VALUE;

  if FAsOpen then
  begin
    FFile := FileOpen(FFileName, fmOpenRead);
    Read;
  end else
  begin
    if not FileExists(FFileName) then
      FFile := FileCreate(FFileName)
    else
      FFile := FileOpen(FFileName, fmOpenReadWrite);
  end;
end;

destructor TWinLayerFS.Destroy;
begin
  if (FFile <> INVALID_HANDLE_VALUE) then
  begin
    if not FAsOpen then
      Write;
    FileClose(FFile);
  end;
  Clear; 

  inherited;
end;

function TWinLayerFS.GetCount: Integer;
begin
  Result := Length(FItems);
end;

function TWinLayerFS.GetItem(Index: Integer): TWinLayerFSItem;
begin
  Result := FItems[Index];
end;

procedure TWinLayerFS.SetItem(Index: Integer; Value: TWinLayerFSItem);
begin
  FItems[Index] := Value;
end;

procedure TWinLayerFS.AddFont(Name, FontName: String; Size: Integer;
  Color: TColor; Style: TFontStyles);
var
  p: PWinLayerFSFont;
begin
  new(p);
  p^.Name := FontName;
  p^.Size := Size;
  p^.Color := Color;
  p^.Style := FontStylesToBin(Style);
  SetLength(FItems, Length(FItems) + 1);
  with FItems[Length(FItems) - 1] do
  begin
    dwName := Name;
    dwType := WinLayerTypeFont;
    dwData := p;
    dwState := True;
  end;
end;

procedure TWinLayerFS.AddInt(Name: String; Value: String);
var
  p: PString;
begin
  new(p);
  p^ := Value;
  SetLength(FItems, Length(FItems) + 1);
  with FItems[Length(FItems) - 1] do
  begin
    dwName := Name;
    dwType := WinLayerTypeInt;
    dwData := p;
    dwState := True;
  end;
end;

procedure TWinLayerFS.AddColor(Name: String; Color: DWORD);
var
  p: PDWORD;
begin
  new(p);
  p^ := Color;
  SetLength(FItems, Length(FItems) + 1);
  with FItems[Length(FItems) - 1] do
  begin
    dwName := Name;
    dwType := WinLayerTypeColor;
    dwData := p;
    dwState := True;
  end;
end;

procedure TWinLayerFS.AddImage(Name: String; Image: TPNGObject);
begin
  SetLength(FItems, Length(FItems) + 1);
  with FItems[Length(FItems) - 1] do
  begin
    dwName := Name;
    dwType := WinLayerTypeImage;
    dwData := Image;
    dwState := Assigned(Image);
  end;
end;

procedure TWinLayerFS.AddImage(Name, FileName: String);
var
  Png: TPNGObject;
begin
  Png := TPNGObject.Create;
  Png.LoadFromFile(FileName);
  AddImage(Name, Png);
end;

procedure TWinLayerFS.Delete(Index: Integer);
var
  i: Integer;
begin
  for i:= Index to Length(FItems) - 2 do
    FItems[i] := FItems[i + 1];
  case FItems[Length(FItems) - 1].dwType of
    WinLayerTypeFont:
      dispose(PWinLayerFSFont(FItems[Length(FItems) - 1].dwData));
    WinLayerTypeInt:
      dispose(PInteger(FItems[Length(FItems) - 1].dwData));
    WinLayerTypeColor:
      dispose(PDWORD(FItems[Length(FItems) - 1].dwData));
    WinLayerTypeImage:
      if FItems[Length(FItems) - 1].dwState then
        TPNGObject(FItems[Length(FItems) - 1].dwData).Free;
  end;
  SetLength(FItems, Length(FItems) - 1);
end;

procedure TWinLayerFS.Clear;
var
  i: Integer;
begin
  for i:= 0 to Length(FItems) - 1 do
  case FItems[i].dwType of
    WinLayerTypeFont:
      dispose(PWinLayerFSFont(FItems[i].dwData));
    WinLayerTypeInt:
      dispose(PInteger(FItems[i].dwData));
    WinLayerTypeColor:
      dispose(PDWORD(FItems[i].dwData));
    WinLayerTypeImage:
      if FItems[i].dwState then
        TPNGObject(FItems[i].dwData).Free;
  end;  
  SetLength(FItems, 0);
end;

procedure TWinLayerFS.Read;
var
  Header: TWinLayerFSHeader;
  Item: TWinLayerFSHeaderDescItem;
  i: Integer;
  Offset: DWORD;
  FontNameLen: WORD;
  Name, TmpW: WideString;
  Buff: array[0..1023] of CHAR;
  CRC: DWORD;
  Color: DWORD;
  Int: String;
  Font: TWinLayerFSFont;
begin
  Clear;
  FileSeek(FFile, 0, 0);
  FileRead(FFile, Header, SizeOf(Header));

  CRC := 0;
  FileSeek(FFile, Header.dwDescOffset, 0);
  while True do
  begin
    i := FileRead(FFile, Buff, SizeOf(Buff));
    if (i <= 0) then
      Break;
    CalcBufCRC32(CRC, @Buff, i);
  end;

  if (Header.dwCRC <> CRC) then
    Exit;   
           
  Offset := Header.dwDescOffset;
  while (Offset < Header.dwNameOffset) do
  begin
    FileSeek(FFile, Offset, 0);
    FileRead(FFile, Item, SizeOf(Item));

    SetLength(Name, Item.dwNameLen);
    FileSeek(FFile, Item.dwNameOffset, 0);
    FileRead(FFile, PWCHAR(Name)^, Item.dwNameLen * 2);

    case Item.dwType of
      WinLayerTypeFont:
        begin
          FileSeek(FFile, Item.dwOffset, 0);
          FileRead(FFile, FontNameLen, SizeOf(FontNameLen));
          SetLength(TmpW, FontNameLen);
          FileRead(FFile, PWCHAR(TmpW)^, FontNameLen * 2);
          Font.Name := WideCharToString(PWCHAR(TmpW));
          FileRead(FFile, Font.Size, SizeOf(Font.Size));
          FileRead(FFile, Font.Color, SizeOf(Font.Color));
          FileRead(FFile, Font.Style, SizeOf(Font.Style));
          AddFont(Name, Font.Name, Font.Size, Font.Color, BinToFontStyles(Font.Style));
          with FItems[Length(FItems) - 1] do
          begin
            dwOffset := Item.dwOffset;
            dwSize := Item.dwSize;
          end;
        end;

      WinLayerTypeInt:
        begin
          FileSeek(FFile, Item.dwOffset, 0);
          SetLength(Int, Item.dwSize);
          FileRead(FFile, Pointer(Int)^, Item.dwSize);

          AddInt(WideCharToString(PWCHAR(Name)), Int);
          with FItems[Length(FItems) - 1] do
          begin
            dwOffset := Item.dwOffset;
            dwSize := Item.dwSize;
          end;
        end;

      WinLayerTypeColor:
        begin
          FileSeek(FFile, Item.dwOffset, 0);
          FileRead(FFile, Color, Item.dwSize);

          AddColor(WideCharToString(PWCHAR(Name)), Color);
          with FItems[Length(FItems) - 1] do
          begin
            dwOffset := Item.dwOffset;
            dwSize := Item.dwSize;
          end;
        end;

      WinLayerTypeImage:
        begin
          AddImage(WideCharToString(PWCHAR(Name)), NIL);
          with FItems[Length(FItems) - 1] do
          begin
            dwState := False;
            dwOffset := Item.dwOffset;
            dwSize := Item.dwSize;
          end;
        end;
    end;

    inc(Offset, SizeOf(Item));
  end;
end;

procedure TWinLayerFS.Write;
var
  Header: TWinLayerFSHeader;
  Item: TWinLayerFSHeaderDescItem;
  i: Integer;
  NameOffset, DataOffset, DescOffset: DWORD;
  Name: WideString;
  NameLen: WORD;
  ImageStr: TMemoryStream;
  Buff: array[0..1023] of CHAR;
begin
  DescOffset := SizeOf(Header);
  NameOffset := DescOffset + DWORD(SizeOf(Item) * Length(FItems));
  DataOffset := NameOffset;
  for i:= 0 to Length(FItems) - 1 do
    inc(DataOffset, Length(FItems[i].dwName) * 2);

  lstrcpy(Header.dwHead, 'WinLayer SKIN');
  Header.dwDescOffset := DescOffset;
  Header.dwNameOffset := NameOffset;
  Header.dwDataOffset := DataOffset;

  for i:= 0 to Length(FItems) - 1 do
  begin
    Item.dwFlag := $FF;
    Item.dwType := FItems[i].dwType;

    case Item.dwType of
      WinLayerTypeFont:
        begin
          Item.dwNameLen := Length(FItems[i].dwName);
          Item.dwNameOffset := NameOffset;
          with PWinLayerFSFont(FItems[i].dwData)^ do
          begin
            NameLen := Length(Name);
            Item.dwSize := Length(Name) * 2 + SizeOf(Size) + SizeOf(Color) + SizeOf(Style);
          end;
          Item.dwOffset := DataOffset;

          FileSeek(FFile, DescOffset, 0);
          FileWrite(FFile, Item, SizeOf(Item));
          inc(DescOffset, SizeOf(Item));

          FileSeek(FFile, NameOffset, 0);
          SetLength(Name, Length(FItems[i].dwName));
          StringToWideChar(FItems[i].dwName, PWCHAR(Name), Length(FItems[i].dwName) * 2);
          FileWrite(FFile, PWCHAR(Name)^, Length(FItems[i].dwName) * 2);
          SetLength(Name, 0);
          inc(NameOffset, Length(FItems[i].dwName) * 2);

          FileSeek(FFile, DataOffset, 0);
          FileWrite(FFile, NameLen, SizeOf(NameLen));
          SetLength(Name, NameLen);
          StringToWideChar(PWinLayerFSFont(FItems[i].dwData)^.Name, PWCHAR(Name), NameLen * 2);
          FileWrite(FFile, PWCHAR(Name)^, NameLen * 2);
          SetLength(Name, 0);

          with PWinLayerFSFont(FItems[i].dwData)^ do
          begin
            FileWrite(FFile, Size, SizeOf(Size));
            FileWrite(FFile, Color, SizeOf(Color));
            FileWrite(FFile, Style, SizeOf(Style));
          end;   
        end;

      WinLayerTypeInt:
        begin
          Item.dwNameLen := Length(FItems[i].dwName);
          Item.dwNameOffset := NameOffset;
          Item.dwSize := Length(PString(FItems[i].dwData)^);
          Item.dwOffset := DataOffset;      

          FileSeek(FFile, DescOffset, 0);
          FileWrite(FFile, Item, SizeOf(Item));
          inc(DescOffset, SizeOf(Item));

          FileSeek(FFile, NameOffset, 0);
          SetLength(Name, Length(FItems[i].dwName));
          StringToWideChar(FItems[i].dwName, PWCHAR(Name), Length(FItems[i].dwName) * 2);
          FileWrite(FFile, PWCHAR(Name)^, Length(FItems[i].dwName) * 2);
          SetLength(Name, 0);
          inc(NameOffset, Length(FItems[i].dwName) * 2);

          FileSeek(FFile, DataOffset, 0);
          FileWrite(FFile, PString(FItems[i].dwData)^[1], Item.dwSize);
        end;

      WinLayerTypeColor:
        begin
          Item.dwNameLen := Length(FItems[i].dwName);
          Item.dwNameOffset := NameOffset;
          Item.dwSize := 4;
          Item.dwOffset := DataOffset;

          FileSeek(FFile, DescOffset, 0);
          FileWrite(FFile, Item, SizeOf(Item));
          inc(DescOffset, SizeOf(Item));

          FileSeek(FFile, NameOffset, 0);
          SetLength(Name, Length(FItems[i].dwName));
          StringToWideChar(FItems[i].dwName, PWCHAR(Name), Length(FItems[i].dwName) * 2);
          FileWrite(FFile, PWCHAR(Name)^, Length(FItems[i].dwName) * 2);
          SetLength(Name, 0);
          inc(NameOffset, Length(FItems[i].dwName) * 2);

          FileSeek(FFile, DataOffset, 0);
          FileWrite(FFile, PDWORD(FItems[i].dwData)^, Item.dwSize);
        end;

      WinLayerTypeImage:
        begin
          ImageStr := TMemoryStream.Create;
          Prepare(i, True);
          TPNGObject(FItems[i].dwData).SaveToStream(ImageStr);
          Prepare(i, False);

          Item.dwNameLen := Length(FItems[i].dwName);
          Item.dwNameOffset := NameOffset;
          Item.dwSize := ImageStr.Size;
          Item.dwOffset := DataOffset;

          FileSeek(FFile, DescOffset, 0);
          FileWrite(FFile, Item, SizeOf(Item));
          inc(DescOffset, SizeOf(Item));

          FileSeek(FFile, NameOffset, 0);
          SetLength(Name, Length(FItems[i].dwName));
          StringToWideChar(FItems[i].dwName, PWCHAR(Name), Length(FItems[i].dwName) * 2);
          FileWrite(FFile, PWCHAR(Name)^, Length(FItems[i].dwName) * 2);
          SetLength(Name, 0);
          inc(NameOffset, Length(FItems[i].dwName) * 2);

          FileSeek(FFile, DataOffset, 0);
          FileWrite(FFile, ImageStr.Memory^, ImageStr.Size);
          ImageStr.Free;
        end;
    end;

    inc(DataOffset, Item.dwSize);
  end;

  Header.dwSize := FileSeek(FFile, 0, 1);
  Header.dwCRC := 0;
  FileSeek(FFile, Header.dwDescOffset, 0);
  while True do
  begin
    i := FileRead(FFile, Buff, SizeOf(Buff));
    if (i <= 0) then
      Break;
    CalcBufCRC32(Header.dwCRC, @Buff, i);
  end;

  FileSeek(FFile, 0, 0);
  FileWrite(FFile, Header, SizeOf(Header));
end;

procedure TWinLayerFS.Prepare(Index: Integer; State: Boolean);
var
  Str: TMemoryStream;
  Png: TPNGObject;
begin
  if FAsOpen then
  if (FItems[Index].dwType = WinLayerTypeImage) and
    (FItems[Index].dwState <> State) then
  begin
    FItems[Index].dwState := State;
    if State then
    begin
      Str := TMemoryStream.Create;
      Str.SetSize(FItems[Index].dwSize);

      FileSeek(FFile, FItems[Index].dwOffset, 0);
      FileRead(FFile, Str.Memory^, FItems[Index].dwSize);

      Png := TPNGObject.Create;
      Png.LoadFromStream(Str);
      FItems[Index].dwData := Png;

      Str.Free;
    end else
    begin
      TPNGObject(FItems[Index].dwData).Free;
      FItems[Index].dwData := NIL;
    end;
  end;
end;

end.
